//! Problem builders for different calibration tasks.

pub mod handeye;
pub mod laserline_bundle;
mod planar_family_shared;
pub mod planar_intrinsics;
pub mod rig_extrinsics;
pub mod scheimpflug_intrinsics;
